/* Copyright  2000 
Michael Gradman and Corwin Joy 

Permission to use, copy, modify, distribute and sell this software and 
its documentation for any purpose is hereby granted without fee, provided 
that the above copyright notice appears in all copies and that both that 
copyright notice and this permission notice appear in supporting documentation. 
Corwin Joy and Michael Gradman make no representations about the suitability 
of this software for any purpose. 
It is provided "as is" without express or implied warranty.
*/ 
// exception class for DB abstractions
// Initial: 9/26/2000 - MG
// Reviewed: 11/12/2000 - CJ
// Edited: 11/14/2000 - MG - made to inherit from RootException
// Edited: 12/19/2000 - MG - added namespaces

#ifndef _DB_EXCEPTION_H
#define _DB_EXCEPTION_H

#include "std_inc.h"
#include "DB_Base.h"

BEGIN_DTL_NAMESPACE

// kludge needed to get around header hell
class DBConnection;
class DBStmt;

extern HENV GetHENV(const DBConnection &conn);
extern HDBC GetHDBC(const DBConnection &conn);
extern HSTMT GetHSTMT(const DBStmt &stmt);

// exception class for DB Abstractions
class DBException : public RootException
{
  public:
	   // raw information that is used to build an error string for ODBC errors
	   struct ODBCErrorInfo
	   {
		  string szSQLState;
		  SDWORD nativeErr;
		  string errMsg;

		  ODBCErrorInfo() : szSQLState(), nativeErr(0), errMsg() { }

		  ODBCErrorInfo(const string &sqlState, const SDWORD &nErr, const string &msg) :
		     szSQLState(sqlState), nativeErr(nErr), errMsg(msg) { }

		  friend ostream &operator<<(ostream &o, const ODBCErrorInfo &info)
		  {
		     o << "SQL Error! " << "SQLSTATE = " << info.szSQLState 
			   << " Native err = " << info.nativeErr << " msg = " << info.errMsg << endl;

			 return o;
		  }

		  operator string() const
		  {
			 ostrstream ostr;
			 ostr << *this << ends;
			 return ostr.str();
		  }
	   };

  private:
	static vector<ODBCErrorInfo> ShowSQLErrors(const DBConnection *pConn, const DBStmt *pStmt)
	{		
	   unsigned char szSQLSTATE[6];
	   SDWORD nErr;
	   unsigned char msg[SQL_MAX_MESSAGE_LENGTH + 200];
	   SWORD cbmsg;

	   DBConnection *conn = const_cast<DBConnection *>(pConn);
	   DBStmt		*stmt = const_cast<DBStmt *>(pStmt);

	   HENV henv = (conn ? GetHENV(*conn) : 0);
	   HDBC hdbc = (conn ? GetHDBC(*conn) : 0);
	   HSTMT hstmt = (stmt ? GetHSTMT(*stmt) : 0);

	   vector<ODBCErrorInfo> errors;

	   if (henv == SQL_NULL_HENV && hdbc == SQL_NULL_HDBC && hstmt == SQL_NULL_HSTMT) return errors;

       while (SQLError(henv, hdbc, hstmt, (unsigned char *)szSQLSTATE, &nErr, msg, sizeof(msg)-1, &cbmsg) == SQL_SUCCESS) 
	   {
		  // MessageBox(NULL, (char *)msg, (char *)szSQLSTATE,MB_OK);
		  // Note: if we're inserting into the database we have to take out
		  // the single quotes containing the msg.

		  ostrstream statestr;
		  statestr << szSQLSTATE << ends;

		  ostrstream msgstr;
		  msgstr << msg << ends;

		  ODBCErrorInfo info(statestr.str(), nErr, msgstr.str());

		  errors.push_back(info);
	   }

	   return errors;
	}

  public:
    const vector<ODBCErrorInfo> sqlErrors; // errors generated by ODBC

	DBException(const string &meth, const string &msg,
			const DBConnection *pConn, const DBStmt *pStmt) : 
			RootException(meth, msg, "DBException"), 
			sqlErrors(ShowSQLErrors(pConn, pStmt))
	{
		// write out msg to error log
	
	}

	virtual const char *what() const throw()
	{
       string rootWhat = RootException::what();
	   ostrstream o;
	   o << rootWhat;

	   if (sqlErrors.size() > 0)
	   {
		  o << "SQL Errors: " << endl;
		  
		  for (size_t i = 0; i < sqlErrors.size(); i++)
		  {
			  o << "(" << i << ") " << sqlErrors[i] << endl;
		  }

	   }
	   o << ends;

	   // this gymnastics is needed so result isn't destroyed
	   // paste these two lines into all what() code
	   whatbuf = o.str();
	   return whatbuf.c_str();
	}

	// gets the true ODBC error code that caused the exception
	// this will be the szSQLState member of the first element
	// of the sqlErrors vector
	// returns empty string if no ODBC error occurred
	pair<string, string> GetODBCError() const
	{
		if (sqlErrors.empty())
			return make_pair(string(""), string(""));
		else
		{
			const ODBCErrorInfo &info = *sqlErrors.begin();
			return make_pair(info.szSQLState, info.errMsg);
		}
	}

};

END_DTL_NAMESPACE
#endif
